from __future__ import annotations

import random
import re
import sys
from typing import TYPE_CHECKING, Any, List, Tuple, Union

from curl_adapter import CurlCffiAdapter
import requests
from bs4 import BeautifulSoup
from requests import utils

from .base import parse_page
from .boxes import ResultList, Thumbnail, _parse_results
from .enums import Status, ThumbnailType
from .errors import ModdbException
from .pages import Member
from .utils import (
    BASE_URL,
    COMMENT_LIMITER,
    GLOBAL_LIMITER,
    GLOBAL_THROTLE,
    LOGGER,
    concat_docs,
    create_login_payload,
    generate_hash,
    generate_login_cookies,
    get,
    get_date,
    get_page_type,
    get_sitearea,
    get_siteareaid,
    join,
    prepare_request,
    raise_for_status,
    ratelimit,
    soup,
    user_agent_list,
)

if TYPE_CHECKING:
    from .boxes import Comment, Tag
    from .enums import WatchType
    from .pages import Engine, Game, Group, Mod, Review, Team


class Message:
    """A single message within a thread.

    Attributes
    -----------
    id : int
        The id of the message
    member : Thumbnail
        A member type thumbnail representing the member
        who sent the message
    timestamp : datetime.datetime
        Message send date
    text : str
        The html text of the message
    """

    def __init__(self, html: BeautifulSoup):
        member = html.find("a", class_="avatar")

        self.id = int(html["id"])
        self.member = Thumbnail(url=member["href"], type=ThumbnailType.member, name=member["title"])
        self.timestamp = get_date(html.find("time")["datetime"])
        self.text = html.find("div", class_="comment").text

    def __repr__(self):
        return f"< Message member={self.member} >"


class Thread:
    """A thread is a conversation between two or more members in which
    one or more messages can be sent.

    Sorting
    --------
        * **id** - when the message was sent, asc is oldest, desc is most recent
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **hasread** - order by whether or the message has been read, asc is unread first, desc is read first
        * **hasreplied** - order by whether or not you have replied to the message, asc us unreplied first, desc is replied first

    Attributes
    -----------
    name : str
        Name of the thread
    id : int
        Id of the thread
    count : int
        The number of messages in this thread
    members : List[Tumbnail]
        All the members participating in the thread. Note: The
        thumbnail of the member who fetched the message will always
        be first and will have their name attribute be 'you' instead
        of the username
    message : List[Message]
        The messages associated to this thread
    """

    def __init__(self, html: BeautifulSoup):
        thread = html.find("div", id="firstmessage")
        header = thread.find("div", class_="normalcorner")

        header_string = header.find("span", class_="heading").string
        header_match = re.match(r"(.*) \(([0-9]*) messages?\)", header_string)
        if header_match is None:
            self.name = header_string
            self.count = 1
        else:
            self.name = header_match.group(1)
            self.count = int(header_match.group(2))

        members = thread.find("p", class_="introduction").strong.find_all("a")
        option = header.find("a", class_=["deleteicon"])
        messages = thread.find("div", class_=["tablecomments"]).find_all("div", recursive=False)

        self.id = int(option["href"][option["href"].index("=") + 1 :])
        self.members = [
            Thumbnail(url=member["href"], name=member.string, type=ThumbnailType.member)
            for member in members
        ]
        self.messages = [Message(message) for message in messages]

    def __repr__(self):
        return f"< Thread name={self.name} count={self.count} >"


class ThreadThumbnail:
    """The thumbnail of a thread, these represent partial objects that
    must be parsed using a client to get the full thread.

    Attributes
    -----------
    name : str
        The name of the thread
    id : int
        The id of the thread
    url : str
        The url for the thread
    last_messager : Thumbnail
        A thumbnail representig the user that sent the
        last message
    timestamp : datetime.datetime
        Datetime of the last message
    content : str
        Content of the last message. Useful for checking
        the last message without having to parse the thread.
    """

    def __init__(self, **kwargs):
        self.name = kwargs.get("name")
        self.url = join(kwargs.get("url"))
        self.last_messager = kwargs.get("last_messager")
        self.timestamp = kwargs.get("timestamp")
        self.content = kwargs.get("content")
        self.id = int(self.url.split("/")[-1])

    def __repr__(self):
        return f"< ThreadThumbnail name={self.name} last_messager={self.last_messager} >"


@concat_docs
class Update(Thumbnail):
    """An update object. Which is basically just a fancy thumbnail with a couple extra attributes and
    methods.

    Attributes
    -----------
    updates : List[Thumbnail]
        A list of thumbnail objects of the things thave have been posted (new files, new images)
    """

    def __init__(self, **attrs):
        super().__init__(**attrs)

        self.updates = attrs.get("updates")
        self._unfollow_url = join(attrs.get("unfollow"))
        self._clear_url = join(attrs.get("clear"))
        self._client = attrs.get("client")

    def __repr__(self):
        return f"<Update name={self.name} type={self.type.name} updates={len(self.updates)}>"

    def clear(self) -> bool:
        """Clears all updates

        Raises
        -------
        ModdbException
            An error has occured while trying to clear the updates for this page

        Returns
        --------
        bool
            True if the updates were successfully cleared
        """
        r = self._client._request("POST", self._clear_url, data={"ajax": "t"})

        return "successfully removed" in r.json()["text"]

    def unfollow(self) -> bool:
        """Unfollows the page. This will also clear the updates

        Raises
        -------
        ModdbException
            An error has occured while trying to unfollow this page

        Returns
        --------
        bool
            True if the page was successfully unfollowed
        """
        r = self._client._request("POST", self._unfollow_url, data={"ajax": "t"})

        return "no longer watching" in r.json()["text"]


@concat_docs
class Request(Thumbnail):
    """A thumbnail with two extra methods used to clear and accept requests.

    Attributes
    -----------
    """

    def __init__(self, **attrs):
        super().__init__(**attrs)

        self._decline = join(attrs.get("decline"))
        self._accept = join(attrs.get("accept"))
        self._client = attrs.get("client")

    def accept(self) -> bool:
        """Accept the friend request.

        Raises
        -------
        ModdbException
            An error has occured while trying to accept the request

        Returns
        --------
        bool
            True if the request was successfully accepted
        """
        r = self._client._request("POST", self._accept, data={"ajax": "t"})

        return "now friends with" in r.json()["text"]

    def decline(self) -> bool:
        """Decline the friend request

        Raises
        -------
        ModdbException
            An error has occured while trying to decline the request

        Returns
        --------
        bool
            True if the page was successfully declined
        """
        r = self._client._request("POST", self._decline, data={"ajax": "t"})

        return "successfully removed" in r.json()["text"]

    def __repr__(self):
        return f"<Request from={self.name}>"


class Client:
    """Login the user to moddb through the library, this allows user to see guest comments and see
    private groups they are part of. In addition, this can be used for a lot of the operation

    Parameters
    -----------
    username : str
        The username of the user

    password : str
        The password associated to that username

    Raises
    -------
    ValueError
        The password or username was incorrect

    Attributes
    ----------
    member : Member
        The member objects this client instance represents
    """

    def __init__(self, username: str, password: str):
        session = requests.Session()
        session.mount("http://", CurlCffiAdapter())
        session.mount("https://", CurlCffiAdapter())
        session.cookies = generate_login_cookies(username, password, session=session)
        self._session = session
        LOGGER.info("Login successful for %s", username)

        self.member = Member(
            soup(self._request("GET", f"{BASE_URL}/members/{username.replace('_', '-')}").text)
        )

    def __repr__(self):
        return f"<Client username={self.member.name} level={self.member.profile.level}>"

    def __enter__(self):
        self._fake_session = sys.modules["moddb"].SESSION
        sys.modules["moddb"].SESSION = self._session

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.modules["moddb"].SESSION = self._fake_session
        delattr(self, "_fake_session")

    @ratelimit(GLOBAL_THROTLE, GLOBAL_LIMITER)
    def _request(self, method, url, **kwargs):
        """Making sure we do our request with the cookies from this client rather than the cookies
        of the library."""
        cookies = utils.dict_from_cookiejar(self._session.cookies)
        headers = {
            **kwargs.pop("headers", {}),
            "User-Agent": random.choice(user_agent_list),
        }

        req = requests.Request(
            method, url, headers=headers, cookies=cookies, data=kwargs.pop("data", {})
        )
        prepped = self._session.prepare_request(req)
        LOGGER.info("Request: %s", prepped.url)

        r = self._session.send(prepped, allow_redirects=kwargs.pop("allow_redirects", True))
        raise_for_status(r)

        return r

    def get_updates(self) -> List[Update]:
        """Get the current updates the user has for models they are subscribed to.

        Returns
        --------
        List[Update]
            List of updates (thumbnail like objects with extra methods and an extra attribute)
        """
        r = self._request("GET", f"{BASE_URL}/messages/updates")
        html = soup(r.text)
        updates = []

        strings = (
            "Mods Watch",
            "Members Watch",
            "Engines Watch",
            "Groups Watch",
            "Games Watch",
        )
        raw = html.find_all("span", string=strings)
        objects = [
            e.parent.parent.parent.find("div", class_="table").find_all("div", recursive=False)
            for e in raw
        ]

        objects_raw = [item for sublist in objects for item in sublist[:-1]]
        for update in objects_raw:
            thumbnail = update.find("a")
            unfollow = update.find("a", title="Stop Watching")["href"]
            clear = update.find("a", title="Clear")["href"]
            updates_raw = update.find("p").find_all("a")

            updates.append(
                Update(
                    name=thumbnail["title"],
                    url=thumbnail["href"],
                    type=get_page_type(thumbnail["href"]),
                    image=thumbnail.img["src"],
                    client=self,
                    unfollow=unfollow,
                    clear=clear,
                    updates=[
                        Thumbnail(name=x.string, url=x["href"], type=get_page_type(x["href"]))
                        for x in updates_raw
                    ],
                    date=get_date(update.find("time")["datetime"]),
                )
            )

        return updates

    def clear_updates(self, category: WatchType) -> bool:
        """Clear all updates for a specific category of
        watched. This method will conduct two requests because
        the clear update request requires a hash generated by
        moddb.

        Parameters
        -----------
        category: WatchType
            The category of watched to clear

        Returns
        --------
        bool
            Whether or not the clear was successful
        """
        hash_page = self._request("GET", f"{BASE_URL}/messages/updates")
        html = soup(hash_page.text)
        pattern = re.compile(rf".*\/clearall\/(.*)\/1\/{category.name}s")
        link = html.find("a", href=pattern)
        if link is None:
            return True

        r = self._request("POST", f'{BASE_URL}{link["href"]}', data={"ajax": "t"})

        return "updates were cleared" in r.json()["text"]

    def get_friend_requests(self) -> List[Request]:
        """Get the current friend requests the user has.

        Returns
        --------
        List[Request]
            List of requests (thumbnail like objects with extra methods)
        """
        r = self._request("GET", f"{BASE_URL}/messages/updates")
        html = soup(r.text)
        requests = []
        raw = html.find("span", string="Friend Requests")
        raw_requests = raw.parent.parent.parent.find("div", class_="table").find_all(
            "div", recursive=False
        )

        for request in raw_requests[:-1]:
            thumbnail = request.find("a")
            accept = request.find("a", title="Accept")["href"]
            decline = request.find("a", title="Decline")["href"]

            requests.append(
                Request(
                    name=thumbnail["title"],
                    url=thumbnail["href"],
                    type=get_page_type(thumbnail["href"]),
                    image=thumbnail.img["src"],
                    client=self,
                    accept=accept,
                    decline=decline,
                    date=get_date(request.find("time")["datetime"]),
                )
            )

        return requests

    def get_watched(self, category: WatchType, page: int = 1) -> ResultList:
        """Get a list of thumbnails of watched items based on the type parameters. Eventually, you'll also be
        able to paginate your mods.

        Parameters
        -----------
        category : WatchType
            The type of watched thing you wanna get (mod, games, engines)
        page : int
            The page number you want to get

        Returns
        --------
        ResultList[Thumbnail]
            List of watched things

        """
        url = f"{BASE_URL}/messages/watching/{category.name}s"
        html = soup(self._request("GET", f"{url}/page/{page}").text)
        results, current_page, total_pages, total_results = _parse_results(html)

        return ResultList(
            results=results,
            url=url,
            current_page=current_page,
            total_pages=total_pages,
            total_results=total_results,
        )

    def tracking(self, page: Union[Mod, Game, Engine, Group, Member]) -> bool:
        """Follow/unfollow this page.

        Parameters
        -----------
        page : Union[Mod, Game, Engine, Group, Member]
            The page you wish to watch/unwatch

        Raises
        -------
        ModdbException
            An error has occured while trying to follow/unfollow the page

        Returns
        --------
        bool
            True if the page has been successfully followed, False if it has been successfully unfollowed
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "watch",
                "sitearea": get_sitearea(page.url),
                "siteareaid": page.id,
            },
            allow_redirects=False,
        )

        return "be notified" in r.json()["text"]

    def like_comment(self, comment: Comment) -> bool:
        """Like a comment, if the comment has already been liked nothing will happen.

        Parameters
        -----------
        comment : Comment
            The comment to like

        Raises
        -------
        ModdbException
            An error has occured while trying to like the comment

        Returns
        --------
        bool
            True if the comment has been successfully liked
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "karmagood",
                "sitearea": "comment",
                "siteareaid": comment.id,
            },
            allow_redirects=False,
        )

        return "successfully issued" in r.json()["text"]

    def dislike_comment(self, comment: Comment) -> bool:
        """Dislike a comment, if the comment has already been disliked nothing will happen.

        Parameters
        -----------
        comment : Comment
            The comment to dislike

        Raises
        -------
        ModdbException
            An error has occured while trying to dislike the comment

        Returns
        --------
        bool
            True if comment has been successfully disliked.
        """
        if not hasattr(comment, "downvote"):
            raise TypeError("Argument must be a Comment-like object")

        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "karmabad",
                "sitearea": "comment",
                "siteareaid": comment.id,
            },
            allow_redirects=False,
        )

        return "successfully issued" in r.json()["text"]

    def membership(self, page: Union[Group, Team]) -> bool:
        """Join/leave a team

        Parameters
        -----------
        page : Union[Group, Team]
            The team/group you want to join. Will not work if you don't have permissions

        Raises
        -------
        ModdbException
            An error has occured while trying to join/leave the group/team

        Returns
        --------
        bool
            True if the group/team has been successfully joined, False if the group/team has been
            successfully left.

        """
        r = self._request(
            "POST",
            f"{BASE_URL}/groups/ajax/members/change/{page.id}",
            data={"ajax": "t"},
            allow_redirects=False,
        )

        return "successfully joined" in r.json()["text"]

    def report(self, page: Any) -> bool:
        """Report a page. This can take any object that has an id and url attribute.

        Parameters
        -----------
        page : Any
            The page to report

        Raises
        -------
        ModdbException
            An error has occured while trying to report the page

        Returns
        --------
        bool
            True if the page has been successfully reported
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "report",
                "sitearea": get_sitearea(page.url),
                "siteareaid": page.id,
            },
            allow_redirects=False,
        )

        return "already reported this content" not in r.json()["text"]

    def unfriend(self, member: Member) -> bool:
        """Unfriend this member if you are friends with them.

        Parameters
        -----------
        member : Member
            The member you wish to unfriend

        Raises
        -------
        ModdbException
            An error has occured trying to unfriend this user

        Returns
        --------
        bool
            True if the user was succesfully unfriended
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/members/ajax/friends/delete/{member.id}",
            data={"ajax": "t"},
            allow_redirects=False,
        )

        return "no longer friends with this member" in r.json()["text"]

    def send_request(self, member: Member) -> bool:
        """Send a friend request to a user. You will not instantly become friends with them,
        they will have to accept the friend request you sent them first.

        Parameters
        -----------
        member : Member
            The member you wish to send a friend request to

        Raises
        -------
        ModdbException
            An error has occured trying to send a friend request to that user

        Returns
        --------
        bool
            True if the user was succesfully sent a friend request
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/members/ajax/friends/add/{member.id}",
            data={"ajax": "t"},
            allow_redirects=False,
        )

        return "friend request has been sent" in r.json()["text"]

    @ratelimit(COMMENT_LIMITER)
    def add_comment(self, page: Any, text: str, *, comment: Comment = None) -> Any:
        """Add a comment to a page.

        Parameters
        -----------
        page : Any
            Must be a moddb.page, the page you wish to add the comment to.
        test : str
            The content of the comment you wish to post
        comment : Optional[Comment]
            If you wish to reply to another comment you must provide the comment
            object for it there.

        Returns
        --------
        Any
            The page's updated object containing the new comment and any other new data that
            has been posted since then
        """
        r = self._request(
            "POST",
            page.url,
            data={
                "formhash": generate_hash(),
                "replyid": comment.id if comment else 0,
                "page": 1,
                "summary": text,
                "comment": "Save comment",
            },
        )

        return page.__class__(soup(r.text))

    def _comment_state_update(self, comment):
        if comment is None:
            raise ModdbException(
                "This comment no longer exists or is no longer on the page it was initially retrieved from."
            )

        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "delete",
                "sitearea": "comment",
                "siteareaid": comment.id,
                "hash": comment._hash,
            },
            allow_redirects=False,
        )

        return r

    def delete_comment(self, comment: Comment) -> bool:
        """This will delete the supplied comment provided you have the correct permissions.
        This is an expensive request because if how moddb works. It needs to make two requests
        in order to get the correct hash to delete the comment. In addition, it may fail if the
        comment has changed location (page number) from what the object says. It is recommended
        to use a newly created comment object that is less than 30 minutes old.

        Parameters
        -----------
        comment : Comment
            The comment to delete

        Raises
        -------
        ModdbException
            An error occured while trying to delete the comment

        Returns
        --------
        bool
            True if the comment was successfully deleted
        """
        with self:
            page = parse_page(comment._url)
            updated_comment = get(page.comments, id=comment.id)

        r = self._comment_state_update(updated_comment)

        return "You have <u>deleted</u> this comment" in r.json()["text"]

    def undelete_comment(self, comment: Comment) -> bool:
        """This will undelete the supplied comment provided you have the correct permissions.
        This is an expensive request because of how moddb works. It needs to make three requests
        in order to get the correct hash to undelete the comment. In addition, it may fail if the
        comment has changed location (page number) from what the object says. It is recommended
        to use a newly created comment object that is less than 30 minutes old.

        Parameters
        -----------
        comment : Comment
            The comment to undelete

        Raises
        -------
        ModdbException
            An error occured while trying to undelete the comment

        Returns
        --------
        bool
            True if the comment was successfully undeleted
        """
        with self:
            page = parse_page(comment._url)
            updated_comment = get(
                page._get_comments_from_url(comment._url, show_deleted=True),
                id=comment.id,
            )

        r = self._comment_state_update(updated_comment)

        return "You have <u>authorized</u> this comment" in r.json()["text"]

    def edit_comment(self, comment: Comment, new_text: str) -> bool:
        """Edit the contents of a comment. You can only edit your comment 120 minutes after it has
        been posted

        Parameters
        -----------
        comment : Comment
            The comment to edit
        new_text : str
            The new content of the comment

        Raises
        -------
        ModdbException
            An error has occured trying to edit the comment

        Returns
        --------
        bool
            True if the comment was successfully edited
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/comment/ajax/post",
            data={"ajax": "t", "id": comment.id, "summary": new_text},
        )

        return "Your comment has been saved" in r.json()["text"]

    def add_review(
        self, page: Any, rating: int, *, text: str = None, has_spoilers: bool = False
    ) -> bool:
        """Rate and review a page. If you rating is below 3 or above 8 you will be asked
        to also provide a review or else the request will not be made. This is also
        used to edit existing reviews.

        Parameters
        -----------
        page : Union[Mod, Game, Engine, Hardware, Software, Platform]
            The page you wish to review
        rating : int
            The rating from 1 to 10
        text : str
            The text review you are giving of this page
        has_spoilers : bool
            Whether or not this review contains spoilers.

        Raises
        -------
        ModdbException
            An error occured trying to review the page.

        Returns
        --------
        bool
            True of the review was successfuly submitted.

        """
        if not (2 < rating < 9) and text is None:
            raise ModdbException("Please include a review to justify such a low/high rating.")

        with self:
            page = parse_page(page.url)

        r = self._request(
            "POST",
            f"{BASE_URL}/reviews/ajax",
            data={
                "ajax": "t",
                "sitearea": get_sitearea(page.url),
                "siteareaid": page.id,
                "hash": page._review_hash,
                "earlyaccess": int(page.profile.status == Status.early_access),
                "rating": rating,
                "summary": text,
                "spoiler": int(has_spoilers),
            },
            allow_redirects=False,
        )

        return "Your rating has been saved" in r.json()["text"]

    def delete_review(self, review: Review) -> bool:
        """Delete your review on the given page. This function will do two requests in order
        to delete your review.

        Parameters
        -----------
        review : Review
            The review you wish to delete

        Raises
        -------
        ModdbException
            An error occured while trying to delete the review

        Returns
        --------
        bool
            True if the review was successfully deleted
        """
        with self:
            hash_review = self.member.get_reviews()[0]

        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/action/",
            data={
                "ajax": "t",
                "action": "delete",
                "sitearea": "reviews",
                "siteareaid": review.id,
                "hash": hash_review._hash,
                "ispd": 1,
            },
            allow_redirects=False,
        )

        return "You have <u>deleted</u> this review." in r.json()["text"]

    def get_threads(
        self,
        query: str = None,
        read: bool = None,
        replied: bool = None,
        sent_items: bool = False,
        sort: Tuple[str, str] = None,
    ) -> List[ThreadThumbnail]:
        """Get all the messages this user has sent or received. This does not return threads you
        have left.

        Parameters
        -----------
        query : Optional[str]
            Optional query to filter messages
        read : Optional[bool]
            True to filter only read message, false to filter unread, None to allow both
        replied : Optional[bool]
            True to filter messages where you are the last message, False for messages
            where another user is the last message, None for both.
        sent_items:
            Get only the threads you have started
        sort : Tuple[str, str]
            Optional sort tuple to order threads

        Returns
        --------
        List[ThreadThumbnail]
            Thread typed thumbnails
        """
        if sent_items:
            url = f"{BASE_URL}/messages/sentitems"
        else:
            url = f"{BASE_URL}/messages/inbox"

        r = self._request(
            "GET",
            url,
            params={
                "filter": "t",
                "kw": query,
                "hasread": int(read) if read is not None else read,
                "hasreplied": int(replied) if replied is not None else replied,
                "sort": f"{sort[0]}-{sort[1]}" if sort is not None else sort,
            },
        )
        html = soup(r.text)

        threads_raw = html.find_all("div", class_=["tabinbox"])[-1].find_all(
            "div", class_=["rowcontent"]
        )
        threads = []
        for thread in threads_raw:
            member = thread.find("span", class_="subheading").find_all("a")[0]
            threads.append(
                ThreadThumbnail(
                    url=thread.a["href"],
                    name=thread.a["title"],
                    last_messager=Thumbnail(
                        type=ThumbnailType.member, name=member.string, url=member["href"]
                    ),
                    timestamp=get_date(thread.find("time")["datetime"]),
                    content=thread.find("div", class_="content").find("p").string,
                )
            )

        return threads

    def parse_thread(self, thread: ThreadThumbnail) -> Thread:
        """Parse a thread thumbnail into a full thread object.

        Parameters
        ----------
        thread : ThreadThumbnail
            The thumbnail to parse

        Returns
        --------
        Thread
            The parsed thread and its messages
        """
        r = self._request("GET", thread.url)

        return Thread(soup(r.text))

    def send_message(self, members: List[Member], name: str, message: str) -> Thread:
        """Send a message and start a thread with one or more members

        Parameters
        ----------
        member : List[Member]
            The members to send the message to and start the
            thread with
        name : str
            The subject of the message
        message : str
            The message to send

        Returns
        --------
        Thread
            The thread started from sending this message
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/compose/",
            data={
                "formhash": generate_hash(),
                "membersto": ",".join(member.name for member in members),
                "name": name,
                "description": message,
                "messages": "Send+message",
            },
        )

        return Thread(soup(r.text))

    def reply_to_thread(self, thread: Union[Thread, ThreadThumbnail], text: str) -> Thread:
        """Add an additional message to an exiting thread

        Parameters
        -----------
        thread : Union[Thread, ThreadThumbnail]
            The thread to add the message to
        text : str
            The text to send

        Returns
        --------
        Thread
            The updated thread containing the new message. It is
            recommended to use this object as it also contains
            a new hash for sending another message
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/inbox/{thread.id}",
            data={"formhash": generate_hash(), "description": text, "messages": "Send+message"},
        )

        return Thread(soup(r.text))

    def add_member_to_thread(self, thread: Union[Thread, ThreadThumbnail], member: Member) -> bool:
        """Add a member to a conversation

        Parameters
        -----------
        thread : Union[Thread, ThreadThumbnail]
            The thread to add add a member to
        member : Member
            The member to add

        Returns
        --------
        bool
            Whether adding the member was succesful. This
            will not update the thread.
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/members/invite/{thread.id}",
            data={"ajax": "t", "username": member.name, "member": "0"},
            allow_redirects=False,
        )

        return "has been successfully added" in r.json()["text"]

    def leave_thread(self, thread: Union[Thread, ThreadThumbnail]) -> bool:
        """Leave a thread, you will not get any more notifications on this thread.

        Parameters
        ----------
        thread : Union[Thread, ThreadThumbnail]
            The thread to leave

        Returns
        --------
        bool
            Whether leaving the thread was successful
        """
        r = self._request(
            "POST",
            f"{BASE_URL}/messages/ajax/delete/",
            data={"ajax": "t", "thread": thread.id},
            allow_redirects=False,
        )

        return "You have successfully deleted the requested thread" in r.json()["text"]

    def mark_all_read(self) -> bool:
        """Mark all threads as read.

        Returns
        --------
        bool
            True if all threads have been marked as read
        """
        r = self._request("POST", f"{BASE_URL}/messages/ajax/markallread", data={"ajax": "t"})

        return "All messages marked as read" in r.json()["text"]

    def _vote_tag(self, tag: Tag, negative: int):
        params = {
            "ajax": "t",
            "tag": tag.name_id,
            "sitearea": get_siteareaid(tag.sitearea),
            "siteareadid": tag.siteareaid,
            "hash": generate_hash(),
            "negative": str(negative),
        }

        resp = self._request("POST", f"{BASE_URL}/tags/ajax/add", data=params)
        return resp.json()["success"]

    def upvote_tag(self, tag: Tag) -> bool:
        """Upvote a tag

        Parameters
        -----------
        tag : Tag
            The tag to upvote

        Returns
        --------
        bool
            Whether the upvote was successful
        """
        return self._vote_tag(tag, 0)

    def downvote_tag(self, tag: Tag) -> bool:
        """Downvote a tag

        Parameters
        -----------
        tag : Tag
            The tag to downvote

        Returns
        --------
        bool
            Whether the downvote was successful
        """
        return self._vote_tag(tag, 1)


class TwoFactorAuthClient(Client):
    """A subclass of client to be used when facing 2FA requirements."""

    def __init__(self, username: str, password: str):
        self.username = username
        self.password = password

        session = requests.Session()
        session.mount("http://", CurlCffiAdapter())
        session.mount("https://", CurlCffiAdapter())
        self._session = session
        self.member: Member = None

        self._2fa_request: requests.Response = None

    def __repr__(self):
        return f"<Client username={self.username}>"

    def login(self) -> bool:
        """Log the user in

        Returns
        --------
        bool
            True if the login was successful, false it the login requires 2FA
        """
        data, resp = create_login_payload(self.username, self.password, self._session)

        req = requests.Request("POST", f"{BASE_URL}/members/login", data=data, cookies=resp.cookies)
        login = self._session.send(prepare_request(req, self._session), allow_redirects=False)

        if "members2faemailhash" in login.text:
            self._2fa_request = login
            return False

        if "freeman" not in login.cookies:
            raise ValueError(f"Login failed for user {self.username}")

        self._session.cookies = login.cookies

        self.member = Member(
            soup(self._request("GET", f"{BASE_URL}/members/{self.username.replace('_', '-')}").text)
        )
        return True

    def submit_2fa_code(self, code: str) -> Member:
        """Submit the 2FA code sent to the user being logged in. Only works if called
        after `login`

        Returns
        --------
        Member
            The logged in member
        """
        if self._2fa_request is None:
            raise ValueError("Call login first")

        html = soup(self._2fa_request.text)
        form = html.find("form", action="https://www.moddb.com/members/login2fa/#membersform")

        data = {
            "rememberme": "1",
            "referer": "/",
            "2faemaildomain": form.find("input", id="members2faemaildomain")["value"],
            "2faemailhash": form.find("input", id="members2faemailhash")["value"],
            "2faemailcode": code,
            "members": "Verify",
        }

        req = requests.Request(
            "POST", f"{BASE_URL}/members/login2fa", data=data, cookies=self._2fa_request.cookies
        )
        login = self._session.send(prepare_request(req, self._session), allow_redirects=False)

        if "freeman" not in login.cookies:
            raise ValueError(f"Login failed for user {self.username}")

        self._session.cookies = login.cookies
        self._2fa_request = None

        self.member = Member(
            soup(self._request("GET", f"{BASE_URL}/members/{self.username.replace('_', '-')}").text)
        )

        return self.member



import logging

import bs4

from ..boxes import PartialTag, Profile, Thumbnail
from ..enums import ArticleCategory, Difficulty, ThumbnailType, TutorialCategory
from ..utils import LOGGER, concat_docs, get_date, get_views, join
from .base import BaseMetaClass


@concat_docs
class Article(BaseMetaClass):
    """This object represents an news article, a tutorial or a feature.

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    category : :class:`.ArticleCategory`
        Type of the article (news, feature)
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)
    game : Union[:class:`.Game`, :class:`.Object`]
        An game object or an object with an id attribute which represents the
        game the article belongs to.

    Sorting
    --------
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **dateup** - order by article date, asc is most recent first, desc is oldest first
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **member** - order by member???
        * **date** - order by upload date, asc is most recent first, desc is oldest first

        Exclusive to tutorials
        * **meta** - sort by difficulty, asc is most difficult, desc is least difficult
        * **subtype** - sort by the area the tutorial covers

    Attributes
    -----------
    category : ArticleCategory
        Whether this article is a news article, a tutorial or a feature
    name : str
        The name of the article
    profile : Profile
        The profile object of the moddb model the article is for (engine, game, mod...). Can be none if it is not
        rattached to anything, such as for site news.
    tags : List[PartialTag]
         A list of partial tags. You can use `get_tags` and then use the name id to get the right one.
    views : int
        Total amount of times this article was viewed
    today : int
        amount of time the article has been viewed today
    intro : int
        The intro/teaser paragraph of the article
    author : Thumbnail
        A member type thumbnail of the member who published the article
    date : datetime.datetime
        The date the article was published
    html : str
        The html of the article
    plaintext : str
        The article text without any html
    summary : str
        plaintext intro to the article
    tutorial_category : TutorialCategory
        If the article category is tutorial, this represents the area the tutorial covers, else it is None
    difficulty : Difficulty
        If the article category is tutorial, this represents how hard the tutorial is.
    """

    def __init__(self, html: bs4.BeautifulSoup):
        try:
            self.name = html.find("span", itemprop="headline").string
        except AttributeError:
            self.name = html.find("span", itemprop="heading").string

        super().__init__(html)

        raw_type = html.find("h5", string="Browse").parent.span.a.string
        self.category = ArticleCategory[raw_type.lower()]

        try:
            raw = html.find("span", string=raw_type[0:-1]).parent.parent.parent.find(
                "div", class_="table tablemenu"
            )
        except AttributeError:
            raw = html.find("span", string=raw_type).parent.parent.parent.find(
                "div", class_="table tablemenu"
            )

        try:
            self.profile = Profile(html)
        except AttributeError:
            LOGGER.info("'%s' has no profile", self.name, exc_info=LOGGER.level >= logging.DEBUG)
            self.profile = None

        try:
            raw_tags = html.find("form", attrs={"name": "tagsform"}).find_all("a")
            self.tags = [
                PartialTag(x.string, join(x["href"]), x["href"].split("/")[-1])
                for x in raw_tags
                if x.string is not None
            ]
        except AttributeError:
            self.tags = []
            LOGGER.info(
                "'%s' '%s' has no tags",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        views_raw = raw.find("h5", string="Views").parent.span.a.string
        self.views, self.today = get_views(views_raw)

        self.intro = html.find("p", itemprop="description").string
        author = html.find("span", itemprop="author").span.a
        self.author = Thumbnail(name=author.string, url=author["href"], type=ThumbnailType.member)

        self.date = get_date(html.find("time", itemprop="datePublished")["datetime"])
        self.html = str(html.find("div", itemprop="articleBody"))
        self.plaintext = html.find("div", itemprop="articleBody").text

        self.summary = html.find("p", class_="introductiontext").string

        if self.category == ArticleCategory.tutorials:
            cat = html.find("span", itemprop="proficiencyLevel").next_sibling.strip()
            self.tutorial_category = TutorialCategory[
                cat.replace("/", "_").replace(" ", "_").lower()
            ]
            self.difficulty = Difficulty[
                html.find("span", itemprop="proficiencyLevel").string.lower()
            ]

    def __repr__(self):
        return f"<Article title={self.name} type={self.category.name}>"


@concat_docs
class Blog(BaseMetaClass):
    """Object used to represent a member blog.

    Filtering
    ----------
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)

    Sorting
    --------
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **dateup** - order by blog date, asc is most recent first, desc is oldest first
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **member** - order by member???
        * **date** - order by upload date, asc is most recent first, desc is oldest first

    Attributes
    -----------
    name_id : str
        The name_id of the member, cannot be changed, it's a mix of the original username lowercased with
        spaces removed and shortened.

    """

    def __init__(self, *, heading, text):
        author = heading.find("span", class_="subheading").a
        self.author = Thumbnail(url=author["href"], name=author.string, type=ThumbnailType.member)

        self.date = get_date(heading.find("span", class_="date").time["datetime"])

        title = heading.div.h4.a
        self.name = title.string
        self.url = join(title["href"])
        self.name_id = self.url.split("/")[0]

        self.html = str(text.content)
        self.plaintext = text.text

    def __repr__(self):
        return f"<Blog title={self.name}>"



import logging

import bs4

from ..enums import SearchCategory
from ..utils import LOGGER, concat_docs
from .base import PageMetaClass
from .mixins import GetGamesMixin


@concat_docs
class Engine(PageMetaClass, GetGamesMixin):
    """A subclass of Page, however, it does not have the files attribute

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    -----------
    released : :class:`.Status`
        The status of the engine (released, unreleased)
    licence : :class:`.Licence`
        The license of the engine
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first

    Attributes
    -----------
    games : List[Thumbnail]
        A list of games suggested on the engine main page.
    """

    def __init__(self, html: bs4.BeautifulSoup):
        super().__init__(html, SearchCategory.engines)
        delattr(self, "files")

        try:
            self.games = self._get_games(html)
        except AttributeError:
            LOGGER.info(
                "Engine '%s' has no games", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.games = []



from __future__ import annotations

from typing import TYPE_CHECKING, List, Tuple, Union

from ..boxes import Tag
from ..utils import BASE_URL, get_page, get_sitearea

if TYPE_CHECKING:
    from ..boxes import ResultList, Thumbnail
    from ..enums import (
        AddonCategory,
        ArticleCategory,
        Difficulty,
        FileCategory,
        Genre,
        HardwareCategory,
        Licence,
        PlayerStyle,
        RSSType,
        Scope,
        SoftwareCategory,
        Status,
        Theme,
        TimeFrame,
        TutorialCategory,
    )
    from ..utils import Object
    from .game import Game


class GetGamesMixin:
    """Abstract class containing the get_games method"""

    def get_games(
        self,
        index: int = 1,
        *,
        query: str = None,
        status: Status = None,
        genre: Genre = None,
        theme: Theme = None,
        scope: Scope = None,
        players: PlayerStyle = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of games for the model. Each page will yield up to 30 games.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the games for.
        query : Optional[str]
            The text to look for in the game names.
        status : Optional[Status]
            The status of the game (unreleased, released, early access, ect...)
        genre : Optional[Genre]
            The genre of the game (fps, tower defense)
        theme : Optional[Theme]
            The theme of the game (fantasy, action)
        scope : Optional[Scope]
            The scope of the game (AAA, indie)
        players : Optional[PlayerStyle]
            Player styles of the game (co-op, singleplayer)
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            List of game like thumbnails that can be parsed individually.
        """
        params = {
            "filter": "t",
            "kw": query,
            "released": status.value if status else None,
            "genre": genre.value if genre else None,
            "theme": theme.value if theme else None,
            "indie": scope.value if scope else None,
            "players": players.value if players else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/games/page/{index}", params=params)


class GetModsMixin:
    """Abstract class containing the get_mod method"""

    def get_mods(
        self,
        index: int = 1,
        *,
        query: str = None,
        status: Status = None,
        genre: Genre = None,
        theme: Theme = None,
        players: PlayerStyle = None,
        timeframe: TimeFrame = None,
        game: Union[Game, Object] = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of mods for the game. Each page will yield up to 30 mods.

        Parameters
        -----------
        index : int
            The page number to get the mods from.
        query : Optional[str]
            The text to look for in the mod names.
        status : Optional[Status]
            The status of the mod (unreleased, released, early access, ect...)
        genre : Optional[Genre]
            The genre of the mod (fps, tower defense)
        theme : Optional[Theme]
            The theme of the mod (fantasy, action)
        players : Optional[PlayerStyle]
            Player styles of the mod (co-op, singleplayer)
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        game : Union[mod, Object]
            An mod object or an object with an id attribute which represents the
            mod the mod belongs to.
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of mods type thumbnails parsed from the game
        """
        params = {
            "filter": "t",
            "kw": query,
            "released": status.value if status else None,
            "genre": genre.value if genre else None,
            "theme": theme.value if theme else None,
            "players": players.value if players else None,
            "timeframe": timeframe.value if timeframe else None,
            "game": game.id if game else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/mods/page/{index}", params=params)


class GetEnginesMixin:
    """Abstract class containing the get_engines method"""

    def get_engines(
        self,
        index: int = 1,
        *,
        query: str = None,
        status: Status = None,
        licence: Licence = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of engines for the game. Each page will yield up to 30 engines.

        Parameters
        -----------
        index : int
            The page number to get the engines from.
        query : Optional[str]
            The text to look for in the engine names.
        status : Optional[Status]
            The status of the game (unreleased, released, early access, ect...)
        licence : Optional[Licence]
            The licence of the engine (open source, proprietary, ect...)
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results
        Returns
        --------
        ResultList[Thumbnail]
            The list of engine type thumbnails parsed from the game
        """
        params = {
            "filter": "t",
            "kw": query,
            "released": status.value if status else None,
            "licence": licence.value if licence else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/engines/page/{index}", params=params)


class SharedMethodsMixin:
    """Abstract class that implements a certain amount of top level methods shared between Pages
    and Hardware"""

    def get_reviews(
        self,
        index: int = 1,
        *,
        query: str = None,
        rating: int = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of reviews for the page. Each page will yield up to 10 reviews.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the reviews from.
        query : Optional[str]
            The string to look for in the review, optional.
        rating : Optiona[int]
            A number between 1 and 10 to get the ratings
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Review]
            The list of reviews parsed from the page
        """
        from .opinion import ReviewList, parse_reviews

        params = {
            "filter": "t",
            "kw": query,
            "rating": rating,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        base_url = f"{self.url}/reviews"
        url = f"{base_url}/page/{index}"
        html = get_page(url, params=params)
        results, current_page, total_pages, total_results = parse_reviews(html)

        return ReviewList(
            results=results,
            url=base_url,
            total_results=total_results,
            params=params,
            current_page=current_page,
            total_pages=total_pages,
        )

    def get_articles(
        self,
        index: int = 1,
        *,
        query: str = None,
        category: ArticleCategory = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of articles for the page. Each page will yield up to 30 articles.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the articles from.
        query : Optional[str]
            The string query to search for in the article name, optional.
        category : Optional[ArticleCategory]
            Type enum defining what the article is, optional
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of article type thumbnails parsed from the page
        """
        params = {
            "filter": "t",
            "kw": query,
            "type": category.value if category else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/articles/page/{index}", params=params)

    def get_files(
        self,
        index: int = 1,
        *,
        query: str = None,
        category: FileCategory = None,
        addon_type: AddonCategory = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of files for the page. Each page will yield up to 30 files.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the files from.
        query : Optional[str]
            The string query to search for in the file name, optional.
        category : [FileCategory]
            Type enum defining what the file is, optional
        addon_type : Optional[AddonCategory]
            Type enum defining what category the file is.
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of file type thumbnails parsed from the page
        """
        params = {
            "filter": "t",
            "kw": query,
            "category": category.value if category else None,
            "categoryaddon": addon_type.value if addon_type else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/downloads/page/{index}", params=params)

    def get_images(self) -> List[Thumbnail]:
        """Get all the images a page has uploaded. Literally all of them. As thumbnails. ModDB's imagebox
        caches all the urls for images on the page so this grabs them all. Lists might be long, but this
        is all the images. They can be invidually parsed to get full fledged media objects from them.

        Returns
        --------
        List[Thumbnail]
            The list of image type thumbnails parsed from the page
        """
        html = get_page(f"{self.url}/images")
        return self._get_media(1, html=html)

    def get_videos(self) -> List[Thumbnail]:
        """Get all the videos a page has uploaded. Literally all of them. As thumbnails. ModDB's imagebox
        caches all the urls for videos on the page so this grabs them all. Lists might be long, but this
        is all the videos. They can be invidually parsed to get full fledged media objects from them.

        Returns
        --------
        List[Thumbnail]
            The list of video type thumbnails parsed from the page
        """
        html = get_page(f"{self.url}/videos")
        return self._get_media(2, html=html)

    def get_tutorials(
        self,
        index: int = 1,
        *,
        query: str = None,
        difficulty: Difficulty = None,
        tutorial_type: TutorialCategory = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of tutorial for the page. Each page will yield up to 30 tutorials.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the tutorials from.
        query : Optional[str]
            The string query to look for in the tutorial title, optional.
        difficulty : Optional[Difficulty]
            Enum type representing the difficulty of the tutorial, optional.
        tutorial_type : Optional[TutorialCategory]
            Enum type representing the theme/type/category of the tutorial, optional.
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of article type thumbnails parsed from the page
        """
        params = {
            "filter": "t",
            "kw": query,
            "subtype": tutorial_type.value if tutorial_type else None,
            "meta": difficulty.value if difficulty else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/tutorials/page/{index}", params=params)


class GetWaresMixin:
    """Abstrac class implementing get_software and get_hardware"""

    def get_hardware(
        self,
        index: int = 1,
        *,
        query: str = None,
        status: Status = None,
        category: HardwareCategory = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of hardware for the platform. Each page will yield up to 30 hardware.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the hardware for.
        query : Optional[str]
            The text to look for in the hardware's names
        status : Optional[Status]
            Status of the hardware (released, unreleased, early access...)
        category : Optional[HardwareCategory]
            Category of the hardware (headset, controller, ect...)
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by

        Returns
        --------
        ResultList[Thumbnail]
            List of hardware like thumbnails that can be parsed individually.
        """
        params = {
            "filter": "t",
            "kw": query,
            "released": status.value if status else None,
            "category": category.value if category else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/hardware/page/{index}", params=params)

    def get_software(
        self,
        index: int = 1,
        *,
        query: str = None,
        status: Status = None,
        category: SoftwareCategory = None,
        timeframe: TimeFrame = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of software for the platform. Each page will yield up to 30 software.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the software for.
        query : Optional[str]
            The text to look for in the hardware's names
        status : Optional[Status]
            Status of the hardware (released, unreleased, early access...)
        category : Optional[SoftwareCategory]
            Category of the hardware (headset, controller, ect...)
        timeframe : Optional[TimeFrame]
            The time period this was released in (last 24hr, last week, last month)
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by

        Returns
        --------
        ResultList[Thumbnail]
            List of software like thumbnails that can be parsed individually.
        """
        params = {
            "filter": "t",
            "kw": query,
            "released": status.value if status else None,
            "category": category.value if category else None,
            "timeframe": timeframe.value if timeframe else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/software/page/{index}", params=params)


class RSSFeedMixin:
    def rss(self, type: RSSType):
        """Get the RSS feed url for the page depending on which feed type you want

        Parameters
        -----------
        type : RSSType
            The type of feed you desire to get

        Returns
        --------
        str
            URL for the feed type
        """
        return f"https://rss.moddb.com/{self._type.name}/{self.name_id}/{type.name}/feed/rss.xml"


class GetAddonsMixin:
    def get_addons(
        self,
        index: int = 1,
        *,
        query: str = None,
        addon_type: AddonCategory = None,
        timeframe: TimeFrame = None,
        licence: Licence = None,
        sort: Tuple[str, str] = None,
    ) -> ResultList:
        """Get a page of addons for the page. Each page will yield up to 30 addons.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the addons from.
        query : Optional[str]
            The string query to search for in the addon name, optional.
        addon_type : Optional[AddonCategory]
            Type enum defining what category the file is.
        timeframe : Optional[TimeFrame]
            Time frame of when the file was added, optional
        licence : Optional[Licence]
            The licence for the addon, optional
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of addon type thumbnails parsed from the page
        """
        params = {
            "filter": "t",
            "kw": query,
            "category": addon_type.value if addon_type else None,
            "timeframe": timeframe.value if timeframe else None,
            "licence": licence.value if licence else None,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }
        return self._get(f"{self.url}/addons/page/{index}", params=params)


class GetWatchersMixin:
    def get_watchers(
        self, index: int = 1, *, query: str = None, sort: Tuple[str, str] = None
    ) -> ResultList:
        """Get a page of watchers for the page. Each page will yield up to 30 members.

        Parameters
        -----------
        index : Optional[int]
            The page number to get the watchers from.
        query : Optional[str]
            The string query to search for in the watcher name, optional.
        sort : Optional[Tuple[str, str]]
            The sorting tuple to sort by the results

        Returns
        --------
        ResultList[Thumbnail]
            The list of member type thumbnails parsed from the page
        """

        params = {
            "filter": "t",
            "kw": query,
            "sort": f"{sort[0]}-{sort[1]}" if sort else None,
        }

        return self._get(f"{self.url}/watchers/page/{index}", params=params)


class GetTagsMixin:
    def get_tags(self):
        """Get more tags for a page.

        Returns
        --------
        List[Tag]
            List of returned tags
        """

        params = {"ajax": "t", "sitearea": get_sitearea(self.url), "siteareaid": self.id}

        resp = get_page(f"{BASE_URL}/tags/ajax/more", params=params, json=True)

        # when there are not extra tags to return resp["tags"]
        # is an empty list instead of a dict
        if isinstance(resp["tags"], list):
            return []

        return [Tag(**tag) for tag in resp["tags"].values()]



# GENERATED AUTOMATICALLY, DO NOT EDIT
import enum


class SearchCategory(enum.Enum):
    """The list of things you can search for"""

    games = 0
    mods = 1
    addons = 2
    downloads = 3
    videos = 4
    articles = 5
    engines = 6
    developers = 7
    groups = 8
    forum = 9
    jobs = 10
    images = 11
    audio = 12
    reviews = 13
    headlines = 14
    blogs = 15
    hardwares = 16
    softwares = 17
    members = 18
    news = articles
    tutorials = articles
    companys = developers
    features = articles


class Category(enum.Enum):
    """Enum for the different areas of the site which things can be attached to"""

    mods = "mods"
    games = "games"
    engines = "engines"
    downloads = "downloads"


class RSSType(enum.Enum):
    """Enum to define the type of RSS you want to get from this page"""

    articles = 0
    downloads = 1
    images = 2
    videos = 3
    tutorials = 4
    reviews = 5
    addons = 6
    blogs = 7
    headlines = 8
    news = 9
    audio = 10
    jobs = 11
    poll = 12


class MediaCategory(enum.Enum):
    """What category a media object is, use for read purposes."""

    video = 0
    image = 1
    audio = 2


class ArticleCategory(enum.Enum):
    """Category of the article"""

    news = 1
    features = 2
    tutorials = 4


class Membership(enum.Enum):
    """Member ship settings of Groups and Teams"""

    invitation = 1
    application = 2
    open_to_all = 3


class TeamCategory(enum.IntFlag):
    """Category of companies, either publisher, developer or both"""

    def __repr__(self):
        cls = self.__class__
        if self._name_ is not None:
            return self._name_
        members, uncovered = enum._decompose(cls, self._value_)
        return "|".join([str(m._name_ or m._value_) for m in members])

    __str__ = __repr__

    developer = 3
    publisher = 4


class ThumbnailType(enum.Enum):
    """The various types of thunbails that can be created"""

    mod = 0
    game = 1
    engine = 2
    member = 3
    group = 4
    article = 5
    review = 6
    team = 7
    blog = 8
    addon = 9
    file = 10
    job = 11
    platform = 12
    media = 13
    software = 14
    hardware = 15
    company = team


class PlayerStyle(enum.IntFlag):
    """The player style of the game"""

    def __repr__(self):
        cls = self.__class__
        if self._name_ is not None:
            return self._name_
        members, uncovered = enum._decompose(cls, self._value_)
        return "|".join([str(m._name_ or m._value_) for m in members])

    __str__ = __repr__

    singleplayer = 1
    multiplayer = 2
    coop = 4
    mmo = 8


class TimeFrame(enum.Enum):
    """How recently the page was updated/uploaded, 24 hours, last week, last month, ect..."""

    day = 1
    week = 2
    month = 3
    year = 4
    more = 5


class WatchType(enum.Enum):
    mod = 0
    game = 1
    engine = 2
    group = 3
    member = 4


class Month(enum.Enum):
    january = "01"
    february = "02"
    march = "03"
    april = "04"
    may = "05"
    june = "06"
    july = "07"
    august = "08"
    september = "09"
    october = "10"
    november = "11"
    december = "12"


# BELOW THIS LINE ENUMS ARE GENERATED AUTOMATICALLY
# PR changes to scripts/generate_enums.py if you want to
# change something


class Difficulty(enum.Enum):
    """Difficulty of the tutorial"""

    basic = 1
    intermediate = 2
    advanced = 3


class TutorialCategory(enum.Enum):
    """Skill covered by the tutorial"""

    coding = 1
    client_side_coding = 16
    server_side_coding = 17
    graphics = 2
    ui_hud = 4
    textures = 5
    concept_art = 3
    level_design = 23
    mapping_technical = 6
    level_design_theory = 24
    modelling = 7
    players_modelling = 8
    weapons_modelling = 9
    props_modelling = 19
    animation = 10
    skinning = 11
    sound = 12
    sound_effects = 13
    voice_acting = 14
    music = 25
    general = 26
    pr = 18
    website = 21
    design_concepts = 31
    management = 15
    starting_a_mod = 32
    installers = 27
    server_tools = 28
    qa_testing = 29
    other = 30


class GroupCategory(enum.Enum):
    """Category of fan groups"""

    official = 1
    arts_literature = 4
    developer_publisher = 99
    educational = 2
    entertainment_press = 9
    event = 11
    fans_clans = 3
    geographic = 7
    hardware_tech = 5
    hobbies_interests = 6
    web_community = 10
    other = 8


class AddonCategory(enum.Enum):
    """Category of addons"""

    maps = 100
    multiplayer_map = 101
    singleplayer_map = 102
    prefab = 103
    models = 104
    npc_model = 105
    player_model = 106
    prop_model = 132
    vehicle_model = 107
    weapon_model = 108
    model_pack = 131
    skins = 110
    npc_skin = 111
    player_skin = 112
    prop_skin = 133
    vehicle_skin = 113
    weapon_skin = 114
    skin_pack = 134
    audio = 116
    music = 117
    npc_audio = 135
    player_audio = 119
    vehicle_audio = 120
    weapon_audio = 121
    ambience_sounds = 137
    language_sounds = 138
    audio_pack = 118
    graphics = 123
    decal = 124
    effects_gfx = 136
    gui = 125
    hud = 126
    spray = 127
    sprite = 128
    texture = 129


class Licence(enum.Enum):
    """The licence of the object"""

    bsd = 7
    commercial = 1
    creative_commons = 2
    gpl = 5
    lgpl = 6
    zlib = 9
    mit = 8
    proprietary = 3
    public_domain = 4


class Status(enum.Enum):
    """Status of a page"""

    released = 1
    early_access = 4
    coming_soon = 2
    unreleased = 3


class Theme(enum.Enum):
    """Theme of the page"""

    abstract = 18
    anime = 1
    antiquity = 14
    comedy = 2
    comic = 3
    education = 17
    fantasy = 4
    fighter = 5
    history = 15
    horror = 6
    mafia = 20
    medieval = 16
    movie = 7
    music = 21
    nature = 13
    noire = 19
    pirate = 22
    postapocalyptic = 25
    realism = 8
    scifi = 9
    survival = 23
    sport = 12
    urban = 24
    war = 10
    western = 11


class Genre(enum.Enum):
    """Genre of the page"""

    action = 1
    first_person_shooter = 3
    third_person_shooter = 34
    tactical_shooter = 4
    fighting = 20
    arcade = 39
    stealth = 43
    adventure = 5
    sub_adventure = 2
    platformer = 19
    point_and_click = 40
    visual_novel = 50
    driving = 6
    racing = 7
    car_combat = 8
    rpg = 10
    role_playing = 11
    roguelike = 44
    hack_n_slash = 51
    party_based = 53
    strategy = 14
    real_time_strategy = 15
    real_time_shooter = 42
    real_time_tactics = 45
    turn_based_strategy = 16
    turn_based_tactics = 46
    tower_defense = 47
    grand_strategy = 49
    x4 = 48
    moba = 52
    sport = 18
    baseball = 13
    basketball = 17
    football = 21
    golf = 33
    hockey = 35
    soccer = 36
    wrestling = 37
    sub_sport = 26
    simulation = 22
    combat_sim = 24
    futuristic_sim = 25
    realistic_sim = 23
    puzzle = 27
    cinematic = 41
    educational = 38
    family = 31
    party = 29
    rhythm = 28
    tcg_ccg_card_game = 54
    virtual_life = 30
    sub_puzzle = 32


class Scope(enum.Enum):
    """Scope of the game"""

    aaa = 1
    indie = 2


class FileCategory(enum.Enum):
    """The category of the File"""

    releases = 1
    full_version = 2
    demo = 3
    patch = 4
    script = 28
    server = 5
    trainer = 29
    media = 6
    trailer = 7
    movie = 8
    music = 9
    audio = 25
    wallpaper = 10
    tools = 11
    audio_tool = 12
    archive_tool = 20
    graphics_tool = 13
    mapping_tool = 14
    modelling_tool = 15
    installer_tool = 16
    server_tool = 17
    plugin = 27
    ide = 18
    sdk = 19
    source_code = 26
    miscellaneous = 21
    guide = 22
    tutorial = 23
    language_pack = 30
    other = 24


class JobSkill(enum.Enum):
    """The skill required for a job"""

    artists = 1
    audio_music = 2
    human_resources = 3
    level_designers = 4
    management = 5
    marketing = 6
    programmers = 7
    public_relations = 8
    qa_testers = 9
    sales = 10
    writers = 12
    web_other = 11


class HardwareCategory(enum.Enum):
    """The category of the hardware"""

    headset = 1
    haptics = 2
    controller = 3
    video = 4
    sound = 5


class SoftwareCategory(enum.Enum):
    """The category of the software"""

    programming = 1
    ide = 2
    sdk = 3
    emulator = 4
    d3_art = 5
    d3_graphics = 6
    voxel_editor = 7
    normal_map_generator = 8
    uv_mapping_and_unwrapping = 9
    heightmap_or_terrain_generator = 10
    shader_editor = 11
    d2_art = 12
    raster_graphics_editor = 13
    vector_graphics_editor = 14
    sprite_editor_and_rigger = 15
    tile_or_map_editor = 16
    sound_and_music = 17
    audio_editing = 18
    digital_audio_workstation = 19
    music_sequencer = 20
    software_synthesizer = 21
    gaming = 22
    gaming_client = 23
    voip = 24
    overclocking = 25
    benchmarking = 26
    streaming_and_recording = 27
    web_pr_and_other = 28
    game_design = 29
    web_design = 30
    browser = 31
    social_networking = 32
    video_editing = 33
    vr = 34



import datetime
import functools
import inspect
import logging
import random
import re
import ssl
import sys
import time
import uuid
from typing import Optional, Sequence, Tuple, TypeVar
from urllib.parse import urljoin

import bs4
import requests
from bs4 import BeautifulSoup, Tag
from requests import utils

from .enums import MediaCategory, ThumbnailType
from .errors import AuthError, AwaitingAuthorisation, ModdbException, Ratelimited

LOGGER = logging.getLogger("moddb")
BASE_URL = "https://www.moddb.com"


time_mapping = {
    "year": 125798400,
    "month": 2419200,
    "week": 604800,
    "day": 86400,
    "hour": 3600,
    "minute": 60,
    "econd": 1,
}

user_agent_list = [
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 Edg/129.0.2792.65",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox One) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 Edge/44.18363.8131",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:130.0) Gecko/20100101 Firefox/130.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0",
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 OPR/114.0.0.0",
    "Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36 OPR/114.0.0.0",
]


def concat_docs(cls):
    """Does it look like I'm enjoying this?"""
    attributes = []

    def get_docs(parent):
        nonlocal attributes
        if parent.__name__ == "object":
            return

        docs = parent.__doc__.splitlines()
        if "    Attributes" in docs:
            attributes = docs[docs.index("    Attributes") + 2 :] + attributes

        source = inspect.getsource(parent.__init__)
        source = source[source.index("):") :]

        if "super().__init__" in source:
            get_docs(parent.__base__)
        elif "__init__" in source:
            get_docs(parent.__base__.__base__)

    get_docs(cls)
    original = cls.__doc__.splitlines()
    if "    Attributes" not in original:
        original.append("    Attributes")
        original.append("    -----------")

    final = original[: original.index("    Attributes") + 2]
    final.extend([x for x in attributes if x.strip()])
    cls.__doc__ = "\n".join(final)

    return cls


class SSLAdapter(requests.adapters.HTTPAdapter):
    def init_poolmanager(self, *args, **kwargs):
        ssl_context = ssl.create_default_context()
        ssl_context.maximum_version = ssl.TLSVersion.TLSv1_2

        kwargs["ssl_context"] = ssl_context
        return super().init_poolmanager(*args, **kwargs)


class Ratelimit:
    def __init__(self, rate: float, per: float, sleep: Optional[None] = None):
        self.rate = rate
        self.per = per
        self.sleep = sleep

        self.last_called = datetime.datetime.min
        self.initial_call = datetime.datetime.min
        self.call_count = 0

    def reset(self, now: datetime.datetime = None):
        if now is None:
            now = datetime.datetime.now()

        self.initial_call = now
        self.call_count = 0

    def call(self):
        now = datetime.datetime.now()

        expiry = self.initial_call + datetime.timedelta(seconds=self.per)
        if now > expiry:
            LOGGER.info("Resetting ratelimit")
            self.reset(now)

        if self.call_count + 1 > self.rate:
            remaining = (expiry - now).total_seconds()
            if self.sleep is not None and remaining <= self.sleep:
                LOGGER.info("Ratelimited! Sleeping for %s", remaining)
                time.sleep(remaining)
                self.reset(now)
            else:
                raise Ratelimited(f"Ratelimited please try again in {remaining}", remaining)

        self.call_count += 1


def ratelimit(*limiters: Ratelimit):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for limiter in limiters:
                limiter.call()

            return func(*args, **kwargs)

        return wrapper

    return decorator


GLOBAL_LIMITER = Ratelimit(40, 300, sleep=300)
GLOBAL_THROTLE = Ratelimit(5, 1, sleep=1)
COMMENT_LIMITER = Ratelimit(1, 60)
LOGIN_LIMITER = Ratelimit(1, 5)


def get_date(d: str) -> datetime.datetime:
    """A helper function that takes a ModDB string representation of time and returns an equivalent
    datetime.datetime object. This can range from a datetime with the full year to
    second to just a year and a month.

    Parameters
    -----------
    d : str
        String representation of a datetime

    Returns
    -------
    datetime.datetime
        The datetime object for the given string
    """
    try:
        return datetime.datetime.strptime(d[:-3] + d[-2:], "%Y-%m-%dT%H:%M:%S%z")
    except ValueError:
        pass

    try:
        return datetime.datetime.strptime(d, "%Y-%m-%d")
    except ValueError:
        pass

    return datetime.datetime.strptime(d, "%Y-%m")


def prepare_request(req: requests.Request, session: requests.Session):
    """Prepared a request with the appropriate cookies"""
    cookies = utils.dict_from_cookiejar(session.cookies)

    if req.cookies is not None:
        req.cookies = {**req.cookies, **cookies}
    else:
        req.cookies = cookies

    if "User-Agent" not in req.headers:
        req.headers["User-Agent"] = random.choice(user_agent_list)

    return session.prepare_request(req)


def raise_for_status(response: requests.Response):
    """Raise any error that could have occured"""
    try:
        text = response.json()
        if text.get("error", False):
            LOGGER.error(text["text"])
            LOGGER.error(response.request.url)
            LOGGER.error(response.request.body)
            raise ModdbException(text["text"])
    except requests.exceptions.JSONDecodeError:
        pass

    response.raise_for_status()

    if (
        "is currently awaiting authorisation, which can take a couple of days while a"
        in response.text.lower()
    ):
        raise AwaitingAuthorisation(
            "This page is still await authorisation and cannot currently be parsed"
        )


@ratelimit(LOGIN_LIMITER)
def generate_login_cookies(username: str, password: str, session: requests.Session = None):
    """Log a user in and return the `freeman` cookie containing the login hash"""
    if session is None:
        session = sys.modules["moddb"].SESSION

    data, resp = create_login_payload(username, password, session)

    req = requests.Request("POST", f"{BASE_URL}/members/login", data=data, cookies=resp.cookies)
    login = session.send(prepare_request(req, session), allow_redirects=False)

    if "members2faemailhash" in login.text:
        raise AuthError("2FA required, use TwoFactorAuthClient")

    if "freeman" not in login.cookies:
        raise ValueError(f"Login failed for user {username}")

    return login.cookies


def create_login_payload(username: str, password: str, session: requests.Session):
    req = requests.Request("GET", f"{BASE_URL}/members/login")
    resp = session.send(prepare_request(req, session))
    resp.raise_for_status()

    html = soup(resp.text)
    form = html.find("form", attrs={"name": "membersform"})

    username_input = form.find("input", id="membersusername")
    botcatcher = form.find("input", type="text", id=False)

    data = {
        "referer": "",
        username_input["name"]: username,
        botcatcher["name"]: "",
        "password": password,
        "rememberme": ["1"],
        "members": "Sign in",
    }

    return data, resp


@ratelimit(GLOBAL_THROTLE, GLOBAL_LIMITER)
def request(req: requests.Request):
    """Helper function to make get/post requests with the current SESSION object.

    Parameters
    -----------
    req : requests.Request
        The request to perform

    Returns
    -------
    requests.Response
        The returned response object

    """
    session: requests.Session = sys.modules["moddb"].SESSION
    prepped = prepare_request(req, session)
    resp = session.send(prepped)

    raise_for_status(resp)
    return resp


def soup(html: str) -> BeautifulSoup:
    """Simple helper function that takes a string representation of an html page and
    returns a beautiful soup object

    Parameters
    -----------
    html : str
        The string representationg of the html to parse

    Returns
    --------
    bs4.BeautifulSoup
        The parsed html
    """

    return BeautifulSoup(html, "html.parser")


def get_page(url: str, *, params: dict = {}, json: bool = False):
    """A helper function that takes a url and returns a beautiful soup objects. This is used to center
    the request making section of the library. Can also be passed a set of paramaters, used for sorting
    and filtering in the search function.

    Parameters
    -----------
    url : str
        The url to get
    params : dict
        A dictionnary of filters and sorting key-value pairs.
    json : Optional[bool]
        Whether the expected response is json, in which case it will not be soup'd

    Returns
    -------
    bs4.BeautifulSoup
        The parsed html
    """
    resp = request(requests.Request("GET", url, params=params))
    if json:
        return resp.json()

    return soup(resp.text)


def get_views(string: str) -> Tuple[int, int]:
    """A helper function that takes a string representation of total something and
    daily amount of that same thing and returns both as a tuple of ints.

    Parameters
    ------------
    string : str
        The string containing the numbers both total and daily

    Returns
    --------
    Tuple[int, int]
        Tuple contains the total views (first element) and the daily views (second element)
    """
    matches = re.search(r"^([0-9,]*) \(([0-9,]*) today\)$", string)
    views = int(matches.group(1).replace(",", ""))
    today = int(matches.group(2).replace(",", ""))

    return views, today


def join(path: str) -> str:
    """Joins a partial moddb url with the base url and returns the combined url

    Parameters
    -----------
    path : str
        the url to join

    Return
    -------
    str
        The full url.

    """
    if not path.startswith(BASE_URL):
        return urljoin(BASE_URL, path)

    return path


def normalize(string: str) -> str:
    """Removes all extra fluff from a stat to get the barebone content.

    Stats usually have extra words like "members" or "visitors" and have command separated integers.

    Parameters
    -----------
    string : str
        The string to clean up

    Returns
    --------
    str
        The cleaned up stat
    """
    return string.replace(",", "").replace("members", "").replace("member", "").strip()


def get_media_type(img: Tag) -> MediaCategory:
    """Determines whether a media is an image, a video or an audio.

    This is somewhat of a brittle method, don't rely on it too much.

    Parameters
    -----------
    img: bsa.Tag
        The image to check

    Returns
    ---------
    MediaCategory
        The category of the media
    """
    if img is None:
        return MediaCategory.audio
    elif img["src"][-8:-5] == ".mp4":
        return MediaCategory.video
    else:
        return MediaCategory.image


def get_page_type(url: str) -> "ThumbnailType":
    """Get the page type based on a url.

    Parameters
    -----------
    url : str
        The url to get

    Return
    -------
    ThumbnailType
        The type of the page
    """
    regex = r"\/((?!page|pages\b)\b\w+)\/"
    type_mapping = {
        "new": "article",
        "feature": "article",
        "tutorial": "article",
        "download": "file",
        "image": "media",
        "audio": " media",
        "video": "media",
    }

    matches = re.findall(regex, url)
    match = matches[-1][0:-1] if matches[0].endswith("s") else matches[0]

    try:
        page_type = ThumbnailType[match]
    except KeyError:
        page_type = ThumbnailType[type_mapping[match]]

    LOGGER.info("%s is type %s", url, page_type)
    return page_type


def ceildiv(a: int, b: int) -> int:
    "Like a // b but rounded up instead of down."
    return -(a // -b)


def get_list_stats(result_box: bs4.BeautifulSoup, per_page: int = 30) -> Tuple[int, int, int]:
    """Get the current page, total pages and total results from
    a result list

    Parameters
    ------------
    result_box: bs4.BeautifulSoup
        The HTML of the result box from a list of results page
    per_page: Optional[int]
        The number of results per page, important for calculations. Defaults
        to 30, doesn't usually need to be touched

    Returns
    --------
    Tuple[int, int, int]
        The stats in order of: number of current page (starting from 1),
        total number of pages (between 1 and X) and the total results.
    """
    stats = re.match(
        r".*\(([0-9,]*) - ([0-9,]*) of ([0-9,]*)\)",
        result_box.find("div", class_="normalcorner")
        .find("div", class_="title")
        .find("span", class_="heading")
        .string,
    )

    if not stats:  # less than a page
        return 1, 1, None

    max_results = int(stats.group(2).replace(",", ""))
    all_results = int(stats.group(3).replace(",", ""))
    max_page = ceildiv(all_results, per_page)
    current_page = ceildiv(max_results, per_page)

    return current_page, max_page, all_results


class Object:
    """A dud objects that will transform every kwarg given into an attribute"""

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


D = TypeVar("D")


def find(predicate, seq: Sequence[D]) -> Optional[D]:
    """A helper to return the first element found in the sequence
    that meets the predicate. For example: ::

        comment = find(lambda comment: comment.author.name == 'SilverElf', mod.comments.flatten())

    would find the first :class:`.Comment` whose author's name is 'SilverElf' and return it.
    If no entry is found, then ``None`` is returned.

    This is different from `filter`_ due to the fact it stops the moment it finds
    a valid entry.

    .. _filter: https://docs.python.org/3.6/library/functions.html#filter

    Parameters
    -----------
    predicate
        A function that returns a boolean-like result.
    seq : iterable
        The iterable to search through.
    """

    for element in seq:
        if predicate(element):
            return element
    return None


def get(iterable: Sequence[D], **attrs) -> Optional[D]:
    r"""A helper that returns the first element in the iterable that meets
    all the traits passed in ``attrs``. This is an alternative for
    :func:`moddb.utils.find`.

    When multiple attributes are specified, they are checked using
    logical AND, not logical OR. Meaning they have to meet every
    attribute passed in and not one of them.

    To have a nested attribute search (i.e. search by ``x.y``) then
    pass in ``x__y`` as the keyword argument.

    If nothing is found that matches the attributes passed, then
    ``None`` is returned.

    Examples
    ---------

    Basic usage:

    .. code-block:: python3

        article = moddb.utils.get(mod.get_articles(), name='Version 3.5 Released')

    Multiple attribute matching:

    .. code-block:: python3

        comment = moddb.utils.get(mod.get_comments(2), content='Test', karma=3)

    Nested attribute matching:

    .. code-block:: python3

        comment = moddb.utils.get(article.get_comments(), author__name='SilverElf', content='Best article ever')

    Parameters
    -----------
    iterable
        An iterable to search through.
    \*\*attrs
        Keyword arguments that denote attributes to search with.
    """

    def predicate(elem):
        for attr, val in attrs.items():
            nested = attr.split("__")
            obj = elem
            for attribute in nested:
                obj = getattr(obj, attribute)

            if obj != val:
                return False
        return True

    return find(predicate, iterable)


def generate_hash():
    return uuid.uuid4().hex


def get_sitearea(url: str) -> str:
    """Get the site area from a url"""
    return url.split("/")[-2]


siteareaid_mapping = {
    "3": "mods",
    "2": "games",
}


def get_siteareaid(key: str):
    """Get the sitearea id from an int"""
    return siteareaid_mapping.get(str(key), "none")


number_mapping = {"k": 1_000, "m": 1_000_000}


def unroll_number(string: str) -> int:
    identifier = string[-1].lower()
    if identifier.isdigit():
        return int(string)

    number = float(string[:-1]) * number_mapping[identifier]
    return int(number)



from curl_adapter import CurlCffiAdapter
import requests

from .base import front_page, login, logout, parse_page, parse_results, rss, search, search_tags
from .client import Client, TwoFactorAuthClient, Thread
from .enums import *
from .pages import *
from .utils import BASE_URL, LOGGER, Object, get_page, request, soup

SESSION = requests.Session()
SESSION.mount("http://", CurlCffiAdapter())
SESSION.mount("https://", CurlCffiAdapter())

__version__ = "0.14.0"

__all__ = [
    "front_page",
    "login",
    "logout",
    "parse_page",
    "parse_results",
    "rss",
    "search",
    "search_tags",
    "Client",
    "TwoFactorAuthClient",
    "Thread",
    "BASE_URL",
    "LOGGER",
    "Object",
    "get_page",
    "request",
    "soup",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from ..enums import SearchCategory
from ..utils import concat_docs
from .base import PageMetaClass
from .mixins import GetAddonsMixin

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


@concat_docs
class Mod(PageMetaClass, GetAddonsMixin):
    """Basically just a subclass of Page

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    released : :class:`.Status`
        The status of the mod (released, unreleased)
    genre : :class:`.Genre`
        The genre of the mod (fps, tower defense)
    theme : :class:`.Theme`
        The theme of the mod (fantasy, action)
    players : :class:`.PlayerStyle`
        Player styles of the mod (co-op, singleplayer)
    timeframe : :class:`.TimeFrame`
        The time period this was released in (last 24hr, last week, last month)
    game : Union[:class:`.Game`, :class:`.Object`]
        An game object or an object with an id attribute which represents the
        game the mod belongs to.

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **game** - order by game???
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first

    """

    def __init__(self, html: BeautifulSoup):
        super().__init__(html, SearchCategory.mods)



import json
import logging
import re

import bs4

from ..boxes import PartialTag, Thumbnail
from ..enums import JobSkill, ThumbnailType
from ..utils import LOGGER, join


class Job:
    """Model representing a job proposed on ModDB

    Parameters
    -----------
    html : bs4.BeautifulSoup
        The html to parse. Allows for finer control.

    Filtering
    ----------
    skill : :class:`.JobSkill`
        The job skill looked for
    earn : :class:`.bool`
        Whether or not the job is paid

    Sorting
    --------
        * **location** - order by the location of the job
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **skill** - order by the skill required
        * **id** - order by upload date, asc is most recent, desc is oldest first

    Attributes
    -----------
    id : int
        id of the job
    name_id : str
        The name_id of the member, cannot be changed, it's a mix of the original username lowercased with
        spaces removed and shortened.
    author : Thumbnail
        A member like thumbnail representing the poster of the job. Can be none if they desire to remain private.
    paid : bool
        Whether or not the job is paid
    tags : List[PartialTag]
        A list of partial tags. You can use `get_tags` and then use the name id to get the right one.
    skill : JobSkill
        the skill demanded for the job
    location : str
        The location the job will be at
    name : str
        The name of the job
    text : str
        The description of the job
    related : List[Thumbnail]
        A list of team like thumbnails of companies related to the job poster

    """

    def __init__(self, html: bs4.BeautifulSoup):
        breadcrumb = json.loads(html.find("script", type="application/ld+json").string)[
            "itemListElement"
        ][-1]["Item"]
        self.name = breadcrumb["name"]
        self.url = breadcrumb["@id"]
        self.name_id = self.url.split("/")[0]
        self.text = html.find("div", id="articlecontent").text

        profile_raw = html.find("span", string="Jobs").parent.parent.parent.find(
            "div", class_="table tablemenu"
        )

        self.id = int(re.search(r"siteareaid=(\d*)", html.find("a", string="Report")["href"])[1])

        try:
            author = profile_raw.find("h5", string="Author").parent.span.a
            self.author = Thumbnail(
                url=author["href"], name=author.string, type=ThumbnailType.member
            )
        except AttributeError:
            LOGGER.info("Job '%s' has no author", self.name, exc_info=LOGGER.level >= logging.DEBUG)
            self.author = None

        self.paid = profile_raw.find("h5", string="Paid").parent.a.string == "Yes"

        try:
            raw_tags = html.find("form", attrs={"name": "tagsform"}).find_all("a")
            self.tags = [
                PartialTag(x.string, join(x["href"]), x["href"].split("/")[-1])
                for x in raw_tags
                if x.string is not None
            ]
        except AttributeError:
            self.tags = []
            LOGGER.info(
                "'%s' '%s' has no tags",
                self.__class__.__name__,
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )

        self.skill = JobSkill(int(profile_raw.find("h5", string="Skill").parent.span.a["href"][-1]))

        self.location = profile_raw.find("h5", string="Location").parent.span.string.strip()

        try:
            related = html.find("div", class_="tablerelated").find_all("a", class_="image")
            self.related = [
                Thumbnail(url=x["href"], name=x["title"], type=ThumbnailType.team) for x in related
            ]
        except AttributeError:
            LOGGER.info(
                "Job '%s' has no related companies",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.related = []

        self._html = html

    def __repr__(self):
        return f"<Job name={self.name}>"



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from ..boxes import Thumbnail
from ..enums import SearchCategory, ThumbnailType
from ..utils import LOGGER, concat_docs
from .base import HardwareSoftwareMetaClass
from .mixins import GetGamesMixin, GetWaresMixin

if TYPE_CHECKING:
    from bs4 import BeautifulSoup


@concat_docs
class Hardware(HardwareSoftwareMetaClass, GetGamesMixin, GetWaresMixin):
    """Represents a moddb Hardware page

    Parameters
    -----------
    html : BeautifulSoup
        The html file to parse, allows for finer control

    Filtering
    ----------
    released : :class:`.Status`
        Release status of the hardware (released, unreleased, early access)
    category : :class:`.HardwareCategory`
        Category of the hardware
    timeframe : :class:`.TimeFrame`
        How long ago the hardware was released

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **category** - sort alphebatically by hardwarecategory
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first

    Attributes
    -----------
    hardware : List[Thumbnail]
        A list of hardware suggested on the hardware main page.
    software : List[Thumbnail]
        A list of software suggested on the hardware main page.
    games : List[Thumbnail]
        A list of games suggested on the hardware main page.
    history : List[Thumbnail]
        A list of previous iterations of the hardware
    recommended : List[Thumbnail]
        A list of recommended hardwares.
    """

    def __init__(self, html: BeautifulSoup):
        super().__init__(html)
        self._type = SearchCategory.hardwares

        try:
            hardware = (
                html.find("span", string="Hardware")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)[:-1]
            )
            self.hardware = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    type=ThumbnailType.hardware,
                    image=x.a.img["src"],
                )
                for x in hardware
            ]
        except AttributeError:
            LOGGER.info(
                "Hardware '%s' has no hardware", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.hardware = []

        try:
            software = (
                html.find("span", string="Software")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)[:-1]
            )
            self.software = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    type=ThumbnailType.software,
                    image=x.a.img["src"],
                )
                for x in software
            ]
        except AttributeError:
            LOGGER.info(
                "Hardware '%s' has no software", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.software = []

        try:
            games = (
                html.find("span", string="Games")
                .parent.parent.parent.find("div", class_="table")
                .find_all("div", recursive=False)[:-1]
            )
            self.games = [
                Thumbnail(
                    name=x.a["title"],
                    url=x.a["href"],
                    type=ThumbnailType.game,
                    image=x.a.img["src"],
                )
                for x in games
            ]
        except AttributeError:
            LOGGER.info(
                "Hardware '%s' has no games", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.games = []

        try:
            history = html.find("span", string="History").parent.parent.parent.find_all(
                "a", class_="image"
            )
            self.history = [
                Thumbnail(
                    url=x["href"],
                    name=x["title"],
                    type=ThumbnailType.hardware,
                    image=x.img["src"],
                )
                for x in history
            ]
        except AttributeError:
            LOGGER.info(
                "Harware '%s' has no history", self.name, exc_info=LOGGER.level >= logging.DEBUG
            )
            self.history = []

        try:
            recommended = html.find("span", string="Recommended").parent.parent.parent.find_all(
                "a", class_="image"
            )
            self.recommended = [
                Thumbnail(
                    url=x["href"],
                    name=x["title"],
                    type=ThumbnailType.hardware,
                    image=x.img["src"],
                )
                for x in recommended
            ]
        except AttributeError:
            LOGGER.info(
                "Hardware '%s' has no recommended",
                self.name,
                exc_info=LOGGER.level >= logging.DEBUG,
            )
            self.recommended = []


@concat_docs
class Software(HardwareSoftwareMetaClass):
    """Represents a moddb Software page

    Parameters
    -----------
    html : BeautifulSoup
        The html file to parse, allows for finer control

    Filtering
    ----------
    released : :class:`.Status`
        Release status of the software (released, unreleased, early access)
    category : :class:`.SoftwareCategory`
        Category of the software
    timeframe : :class:`.TimeFrame`
        How long ago the software was released

    Sorting
    --------
        * **released** - when the object was released, asc is oldest, desc is most recent
        * **id** - when it was added to moddb, asc is oldest, desc is most recent
        * **ranktoday** - order by daily ranking, asc is highest ranked, desc is lowest rank
        * **visitstotal** - order by most views, asc is highest views, desc is lowest views
        * **rating** - order by rating, asc is highest rating, desc is lowest rating
        * **category** - sort alphebatically by softwarecategory
        * **name** - order alphabetically, asc is a-z, desc is z-a
        * **dateup** - order by latest update, asc is most recent update first, desc is oldest update first

    """

    def __init__(self, html: BeautifulSoup):
        super().__init__(html)
        self._type = SearchCategory.softwares
